#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["construct_topic", "extract_topic"]


def construct_topic(value: str, class_name: str) -> tuple[str, str]:
    """Constructs a name and qualified topic.

    :param value: The name of the topic.
    :type value: str
    :param class_name: Name of the calling class to raise sensible error.
    :type class_name: str
    """
    if value.endswith("."):
        raise ValueError(f"{class_name} name cannot end with '.'")
    if value.startswith("."):
        value = value[1:]
        topic = value
    else:
        topic = f"-.{value}" if value else "-"
    return value, topic


def extract_topic(value: str) -> str:
    """Extracts topic from qualified topic name.

    This cuts off a possible ``-.`` at the beginning of the topic name.

    :param value: The qualified name of the topic.
    :type value: str
    """
    return value[2:] if value.startswith("-") else value



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

#  Handlers (listeners) are attached using pypubsub, but we want to isolate this
#  from other usages.

from pubsub.core import Publisher as _Publisher

__all__ = ["send_message", "subscribe"]

_publisher = _Publisher()

send_message = _publisher.sendMessage
subscribe = _publisher.subscribe




#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABCMeta, abstractmethod
from collections import defaultdict
from typing import NamedTuple, Union

from .colors import colorize_string
from .pub_sub import subscribe
from .subject import FrameInformation
from .topic import construct_topic
from .severity import *

__all__ = ['Identifier', 'Handler']


class Identifier(NamedTuple):
    identifier: str
    """.. include:: identifier.txt"""
    continuation: str
    """Only the indicator, indented for use in additional lines."""


class Handler(metaclass=ABCMeta):
    """Ancestor of every concrete Handler.

    A handler processes a possible log entry and adds it to the respective log.

    .. warning:: This class **must** be the last in the list of inherited classes to allow correct processing via Mixin classes.

    :param name: Handler name. When this is an empty string (the default), it signifies this as
        a ROOT Handler. Providing a ``name`` will use it as (sub-)topic, except when the ``name``
        starts with a dot (```.``), then the ``name`` (excluding the dot) will be considered a separate root.
    :type name: str, optional
    """

    severity: int = 0
    """The minimum severity that the Handler will process."""
    date_format: str = "%y-%m-%d %H:%M:%S"
    """Date format used for default Identifier. UTC is used."""
    colors = defaultdict(lambda: "")

    def __init__(self, name: str = "", **kwargs):
        self.name, self.topic = construct_topic(name, self.__class__.__name__)
        subscribe(self, self.topic)

    def __call__(self, subject):
        """Called when a log entry is processed.

        If message is ``None`` this is a configuration message the respective information should available as directory items.
        """
        if subject.message is None and subject.message_id is None:
            # this is a config message to set the severity or something else
            self.config(subject=subject)
        elif subject.traceback:
            self.handle_exception(subject=subject)
        elif subject.severity >= self.severity:
            self.process(subject=subject)

    # stubs that MUST be implemented by concrete classes

    @abstractmethod
    def render(self, identifier: Identifier, content: str, color: str = "") -> str:
        """Render the provided content for output.

        :param identifier: The identifier that will be used.
        :type identifier: :class:`Identifier`
        :param content: The content to be rendered. Might contain new-line(s).
        :type content: str
        :param color: The color of the content if applicable.
        :type content: str
        :return: The rendered content.
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def write(self, content: str):
        raise NotImplementedError()

    # stubs that MIGHT be implemented by concrete classes

    def config(self, subject):
        """Process configuration changes."""
        if "set_severity" in subject:
            self.severity = subject["set_severity"]

    def process(self, subject):
        color = self.colors[subject.severity]
        content_lines = [colorize_string(subject.message, color)] if subject.severity >= CRITICAL else [subject.message]
        if subject.severity in (ERROR, CRITICAL):
            content_lines.append(subject.source_location)
        if subject.severity >= CRITICAL:
            content_lines.append(
                self.format_frame_information(FrameInformation(subject.source.copy(), subject.local_vars.copy()))
            )

        content = self.render(
            identifier=self.get_identifier(subject=subject), content="\n".join(content_lines), color=color
        )
        self.write(content)

    # noinspection PyMethodMayBeStatic
    def format_frame_information(self, frame_information: FrameInformation) -> str:
        """Render Source and local variables for CRITICAL and Exceptions.

        :param frame_information:
        :type frame_information: :class:`hakisto.FrameInformation`
        """
        if not any([frame_information.source, frame_information.local_vars]):
            return ""

        # noinspection PyTypeChecker
        lines: list[tuple[Union[int, str], str]] = frame_information.source.copy()
        num_width = len(str(lines[-1][0]))
        txt_width = max((len(i[1]) for i in lines))

        try:
            var_width = max((len(i) for i in frame_information.local_vars))
        except ValueError:
            var_width = 1
        var_lines = [f"   {k:{var_width}} = {v!r}" for k, v in frame_information.local_vars.items()]

        while len(lines) < len(var_lines):
            lines.insert(0, ("", ""))
        while len(lines) > len(var_lines):
            var_lines.insert(0, "")

        return "\n".join(
            [" " * num_width + " ┌──" + "─" * txt_width + "┐"]
            + [f"{lines[i][0]:{num_width}} │ {lines[i][1]:{txt_width}} │{var_lines[i]}" for i in range(len(lines))]
            + [" " * num_width + " └──" + "─" * txt_width + "┘"]
        )

    def handle_exception(self, subject) -> None:
        """Renders the full traceback with source and local variables.

        :param subject: The subject being processed.
        :type subject: :class:`hakisto.Subject`
        """
        color = self.colors[CRITICAL]

        content_lines = [colorize_string(subject.message, color)]
        for i in subject.traceback:
            content_lines.append(i.source_location)
            content_lines.append(self.format_frame_information(i.frame_information))

        content = self.render(
            identifier=self.get_identifier(subject=subject, indicator="X"),
            content="\n".join(content_lines),
            color=color,
        )
        self.write(content)

    def get_identifier(self, subject, indicator=None):
        """Return standard Identifier

        .. include:: identifier.txt

        :param subject: The subject to be handled
        :type subject: hakisto.Subject:class:`hakisto.Subject`
        :param indicator: Indicator to use instead of the one determined by Subject.severity
        :type indicator: str
        :returns: Identifier
        :rtype: :class:`hakisto.Identifier`
        """
        indicator = indicator or severities.get(subject.severity, str(subject.severity))[0]
        parts = [f"{indicator} {subject.created.strftime(self.date_format)}"]
        if subject.message_id:
            parts.append(f"<{subject.message_id}>")
        if subject.process_name != "MainProcess":
            parts.append(
                f"*{subject.process_name}:{subject.process_id}*" if subject.process_id else f"*{subject.process_name}*"
            )
        if subject.thread_name != "MainThread":
            parts.append(
                f"({subject.thread_name}:{subject.thread_id})" if subject.thread_id else f"({subject.thread_name})"
            )
        if subject.asyncio_task_name:
            parts.append(f"'{subject.asyncio_task_name}'")
        if subject.topic:
            parts.append(subject.topic)
        identifier = f"[{' '.join(parts)}]"
        return Identifier(identifier, f"{' ' * (len(identifier) - len(indicator) - 2)}[{indicator}]")



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# default logger (batteries included)

import inspect as _inspect

from ._logger import Logger as _Logger
from .file_handler import FileHandler
from .stderr_handler import StdErrHandler
from .severity import TRACE, DEBUG, VERBOSE, INFO, SUCCESS, WARNING, ERROR, CRITICAL

_Logger.register_excluded_source_file(_inspect.currentframe().f_code.co_filename)

_logger = _Logger()

critical = _logger.critical
error = _logger.error
warning = _logger.warning
success = _logger.success
info = _logger.info
verbose = _logger.verbose
debug = _logger.debug
trace = _logger.trace
log = _logger.log
set_severity = _logger.set_severity
set_handler_severity = _logger.set_handler_severity

set_severity(INFO)

console_handler = StdErrHandler()
file_handler = FileHandler()



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


from .handler import Identifier
from .colors import colorize_string

__all__ = ['SingleLineMixin']


class SingleLineMixin:
    """Mixin creating a single line, replacing ``\n`` with ``<br />``.

    .. Warning:: Inherit **before** Handler.
    """

    def render(self, identifier: Identifier, content: str, color: str = "") -> str:
        """Render the provided content for output.

        :param identifier: The identifier that will be used.
        :type identifier: :class:`Identifier`
        :param content: The content to be rendered. Might contain new-line(s).
        :type content: str
        :param color: The color of the content if applicable.
        :type content: str
        :return: The rendered content.
        :rtype: str
        """
        content = content.replace("\n", "<br />")
        return f"{colorize_string(identifier.identifier, color)} {content}"



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os

from .indented_mixin import IndentedMixin
from .handler import Handler
from .severity import dark_severity_colors, severity_colors
from .stream import Stream

__all__ = ["StdErrHandler"]


class StdErrHandler(IndentedMixin, Handler):
    """Color Output to console

    :param name: Handler name (used to determine topic)
    :type name: str
    :param colors: Color palette to use
    :type colors: dict[int, str]
    """

    def __init__(self, name: str = "", colors: dict[int, str] = None, **kwargs):
        super().__init__(name, **kwargs)
        if colors is None:
            if os.getenv("HAKISTO_COLORS") and os.getenv("HAKISTO_COLORS").upper() == "DARK":
                colors = dark_severity_colors
            else:
                colors = severity_colors
        self.colors = colors
        self.stream = Stream("<stderr>")

    def write(self, content: str) -> None:
        """Write content to console"""
        with self.stream:
            self.stream.write(f"{content}\n")



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Subject - keeps all information about the logged thing

import asyncio
import datetime
import inspect
import multiprocessing
import threading
from collections.abc import MutableMapping
from types import FrameType, TracebackType
from typing import Optional, Union, NamedTuple, Any, Iterator

from .topic import extract_topic

__all__ = [
    "Subject",
    "SourceExtractLine",
    "FrameInformation",
    "TracebackRecord",
    "get_frame_information",
    "get_traceback_information",
    "get_source_location",
]


class SourceExtractLine(NamedTuple):
    number: int
    """Line number in source file."""
    line: str
    """Line in source file."""


class FrameInformation(NamedTuple):
    source: list[SourceExtractLine] = [].copy()
    """List of :class:`hakisto.SourceExtractLine`."""
    local_vars: dict[str, str] = {}.copy()
    """The local variables as a dictionary with the values being converted to strings."""


class TracebackRecord(NamedTuple):
    source_location: str
    """.. include:: source_location.txt"""
    frame_information: FrameInformation
    """see :class:`hakisto.FrameInformation`."""


def get_frame_information(frame: FrameType) -> FrameInformation:
    """Extract source and local variables from a frame.

    :param frame: Frame object.
    :type frame: :class:`types.FrameType`
    :return: Source and local variables.
    :rtype: :class:`hakisto.FrameInformation`
    """
    try:
        lines, first_line_number = inspect.getsourcelines(frame)
    except OSError:  # pragma: no cover
        return FrameInformation()
    if not first_line_number:  # pragma: no cover
        first_line_number = 1  # for <module> the whole source is returned,
        lines = lines[: frame.f_lineno]  # but the line numbers are wrong
    return FrameInformation(
        [SourceExtractLine(n + first_line_number, line.rstrip()) for n, line in enumerate(lines)],
        {k: str(frame.f_locals[k]) for k in sorted(frame.f_locals) if not k.startswith("__")},
    )


def get_traceback_information(frame: TracebackType) -> list[TracebackRecord]:
    """Return the traceback items, each having source location, source extract, and local variables"""
    result = []
    while frame:
        result.append(TracebackRecord(get_source_location(frame.tb_frame), get_frame_information(frame.tb_frame)))
        frame = frame.tb_next
    return result


def get_source_location(frame: FrameType) -> str:
    """.. include:: source_location.txt"""
    try:
        return f'File "{frame.f_code.co_filename}", line {frame.f_lineno} in {frame.f_code.co_name}'
    except AttributeError:
        return ""


class Subject(MutableMapping):
    """The Subject represents the information that is being logged.

    :param topic: The topic of the subject. A possible root indicator will be removed.
    :param severity: The severity of the entry. A Handler uses the severity to determine if an entry should be made in the respective log.
    :type severity: int
    :param frame: The frame (of the call-stack) that issued the entry. For an Exception this contains the Traceback.
    :type frame: :class:`FrameType`
    :param message: The message of the entry.
    :type message: Optional[str]
    :param message_id: The message id.
    :type message_id: Optional[str]

    Any additional keyword arguments will be available as mapping items.
    """

    severity: int
    """This should be one of the severities below.
    
    .. include:: severities_table.txt
    """
    created: datetime.datetime
    """The **UTC** date and time the logging subject was created"""
    message_id: str
    """Message id, possible coming from a different system that uses these to identify the message reason."""
    thread_name: str
    thread_id: int
    process_name: str
    process_id: int
    asyncio_task_name: str
    topic: str
    """Qualified topic. Starts with ``_`` when in the regular tree."""
    message: str

    frame: Union[FrameType, TracebackType]
    """The execution frame that issued the entry. 
    Used to determine additional (beneficial) information (see items below).
    When the Subject is created be the automatic Exception Handler it points to the respective **traceback**"""

    def __init__(
        self,
        topic: str,
        severity: int,
        frame: FrameType | TracebackType,
        message: str = None,
        message_id: str = None,
        **kwargs,
    ) -> None:
        self.severity = severity
        self.created = datetime.datetime.now(datetime.UTC)
        self.message_id = message_id
        self.message = message
        self.frame = frame

        self.topic = extract_topic(topic)

        self.__data = kwargs.copy()
        self.__cache = {}

        self.thread_name = threading.current_thread().name
        self.thread_id = threading.current_thread().ident
        self.process_name = multiprocessing.current_process().name
        self.process_id = multiprocessing.current_process().ident
        try:
            self.asyncio_task_name = asyncio.current_task().get_name()
        except RuntimeError:
            self.asyncio_task_name = ""

    @property
    def source_location(self) -> str:
        """.. include:: source_location.txt"""
        return get_source_location(self.frame)

    @property
    def source(self) -> list[SourceExtractLine]:
        """Source extract related to the entry."""
        self.__cache_source_and_local_vars()
        return self.__cache["source"]

    @property
    def local_vars(self) -> dict[str, str]:
        """Local variables related to the entry."""
        self.__cache_source_and_local_vars()
        return self.__cache["local_vars"]

    @property
    def traceback(self) -> list[TracebackRecord]:
        """List of traceback items to the entry."""
        if "traceback" not in self.__cache:
            if isinstance(self.frame, TracebackType):
                self.__cache["traceback"] = get_traceback_information(self.frame)
            else:
                self.__cache["traceback"] = []
        return self.__cache["traceback"]

    def __cache_source_and_local_vars(self) -> None:
        if "source" not in self.__cache:
            if isinstance(self.frame, FrameType):
                (self.__cache["source"], self.__cache["local_vars"]) = get_frame_information(self.frame)
            else:
                self.__cache["source"], self.__cache["local_vars"] = tuple(), {}

    def __setitem__(self, key, value, /):
        self.__data[key] = value

    def __delitem__(self, key, /):
        del self.__data[key]

    def __getitem__(self, key, /) -> Any:
        return self.__data[key]

    def __len__(self) -> int:
        return len(self.__data)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__data)

    def __str__(self) -> str:
        return (
            f"{self.severity}, "
            f"{self.created}, "
            f"{self.message_id}, "
            f"{self.thread_name}, "
            f"{self.thread_id}, "
            f"{self.process_name}, "
            f"{self.process_id}, "
            f"{self.asyncio_task_name}, "
            f"{self.topic}, "
            f"{self.message}, "
            f"{self.source_location}, "
            f"{self.__data}"
        )



import sys

from hakisto import logger


def color():
    logger.file_handler.severity = sys.maxsize
    logger.set_severity(0)

    logger.trace('Trace')
    logger.debug('Debug')
    logger.verbose('Verbose')
    logger.info('Info')
    logger.success('Success')
    logger.warning('Warning')
    logger.error('Error')


def critical():
    s = 'Hakisto'
    msg = f'Colors with {s} is easy!'
    logger.critical('A critical issue!')


def traceback():
    s = 'Hakisto'
    msg = f'Colors with {s} is easy!'
    b = s[10]



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# ANSI colors for console output

# fmt: off
BLACK         = '\033[30m'
RED           = '\033[31m'
GREEN         = '\033[32m'
YELLOW        = '\033[33m'
BLUE          = '\033[34m'
MAGENTA       = '\033[35m'
CYAN          = '\033[36m'
WHITE         = '\033[37m'
RESET         = '\033[39m'
LIGHT_BLACK   = '\033[90m'
LIGHT_RED     = '\033[91m'
LIGHT_GREEN   = '\033[92m'
LIGHT_YELLOW  = '\033[93m'
LIGHT_BLUE    = '\033[94m'
LIGHT_MAGENTA = '\033[95m'
LIGHT_CYAN    = '\033[96m'
LIGHT_WHITE   = '\033[97m'
# fmt: on


def colorize_string(value: str, color: str = "") -> str:
    """Return colorized string when a color is provided.

    :param value: The string to colorize.
    :type value: str
    :param color: The color, should be the complete ANSI escape code (see :doc:`colors`).
    :type color: Optional[str]
    :return: The colorized string
    :rtype: str
    """
    if not color:
        return value
    return f"{color}{value}{RESET}"



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from ._logger import *
from .colors import *
from .file_handler import *
from .handler import *
from .indented_mixin import *
from .severity import *
from .single_line_mixin import *
from .stderr_handler import *
from .stream import *
from .subject import *
from .topic import *





#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

#  Handlers (listeners) are attached using pypubsub, but we want to isolate this
#  from other usages.

from types import MappingProxyType
from .colors import *

__all__ = [
    "TRACE",
    "DEBUG",
    "VERBOSE",
    "INFO",
    "SUCCESS",
    "WARNING",
    "ERROR",
    "CRITICAL",
    "severities",
    "severity_colors",
    "light_severity_colors",
    "dark_severity_colors",
]

# fmt: off
TRACE    = 100  #: More information than regular DEBUG (e.g. variable values).
DEBUG    = 200
"""Debugging (technical) information."""
VERBOSE  = 300
"""Additional (non-technical) information."""
INFO     = 400
"""Information."""
SUCCESS  = 500
"""Successful execution."""
WARNING  = 600
"""Warning that might require attention."""
ERROR    = 800
"""Significant issue. Processing might continue."""
CRITICAL = 900
"""Critical issue. Usually processing is aborted."""

severities = MappingProxyType({
    TRACE:    'TRACE',
    DEBUG:    'DEBUG',
    VERBOSE:  'VERBOSE',
    INFO:     'INFO',
    SUCCESS:  'SUCCESS',
    WARNING:  'WARNING',
    ERROR:    'ERROR',
    CRITICAL: 'CRITICAL',
})
"""Lookup severity name using integer value."""

severity_colors = MappingProxyType({
    TRACE:    LIGHT_MAGENTA,
    DEBUG:    LIGHT_CYAN,
    VERBOSE:  LIGHT_GREEN,
    INFO:     LIGHT_GREEN,
    SUCCESS:  LIGHT_GREEN,
    WARNING:  LIGHT_YELLOW,
    ERROR:    LIGHT_RED,
    CRITICAL: LIGHT_RED,
})
"""Default colors for severities."""

light_severity_colors = severity_colors
"""Default light colors for severities."""

dark_severity_colors = MappingProxyType({
    TRACE:    MAGENTA,
    DEBUG:    CYAN,
    VERBOSE:  GREEN,
    INFO:     GREEN,
    SUCCESS:  GREEN,
    WARNING:  YELLOW,
    ERROR:    RED,
    CRITICAL: RED,
})
"""Default dark colors for severities."""
# fmt: on



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# thread save Stream

import inspect
import pathlib
import sys
import threading

from types import FrameType

__all__ = ["Stream", "get_top_caller"]


def get_top_caller() -> FrameType:
    """Return the frame at top of stack, basically the called script."""
    frame = inspect.currentframe()
    while frame.f_back:
        frame = frame.f_back
    return frame


class Stream:
    """Thread save stream, a singleton per ``name`` to prevent intermixing of output when threading is used.

    - If the name is provided will be considered the file name (of the logfile), except
      - ``<stderr>`` will use **stderr**
      - ``<stdout>`` will use **stdout**

    - If the name is **not** provided, the name of the called script will be used (with the extension ``.log``).

    - If the name is **not** provided **and** an interactive session is used, the logfile is named ``__stdin__.log``

    :param name: Name of the stream (logfile)
    :type name: str
    """

    __cache = {}

    def __new__(cls, name: str, *args, **kwargs):
        if name in cls.__cache:
            return cls.__cache[name]
        return super().__new__(cls)

    def __init__(self, name: str, *args, **kwargs):
        self.name = name
        self._stream = None
        self._lock = threading.RLock()
        self._path = None
        if self.name in ("<stdout>", "<stderr>"):
            self.get_stream = lambda: getattr(sys, name[1:-1])
        else:
            if not self.name:
                self.name = get_top_caller().f_code.co_filename
                if self.name == "<stdin>" or self.name.endswith("pydevconsole.py"):
                    self._path = pathlib.Path.cwd().with_name("__stdin__.log")
                else:
                    self._path = pathlib.Path(self.name).with_suffix(".log")
            else:
                self._path = pathlib.Path(self.name)
            self.get_stream = lambda: self.path.open(mode="a", encoding="utf-8")
        self.__cache[self.name] = self
        if self.name.startswith("<std"):
            self.name = ""

    @property
    def path(self) -> pathlib.Path:
        """The file path of the stream. Might be None."""
        return self._path

    @property
    def lock(self):
        """The treading lock. Can be used in a ``with`` statement."""
        return self._lock

    def __enter__(self):
        self._lock.acquire()
        self._stream = self.get_stream()
        return self._stream

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self._stream, "flush"):
            self._stream.flush()
        if self.name and hasattr(self._stream, "close"):
            self._stream.close()
        self._lock.release()

    def open(self):
        """Open the stream and return it."""
        if not self._stream:
            self._stream = self.get_stream()
        return self._stream

    def close(self):
        """Close the stream."""
        if self._stream:
            if all((self.name, hasattr(self._stream, "close"), not self._stream.closed)):
                with self._lock:
                    self._stream.close()
                    self._stream = None

    def flush(self):
        """Flush the stream."""
        if self._stream:
            with self._lock:
                self._stream.flush()

    def write(self, data):
        """Write to the stream."""
        if self._stream:
            with self._lock:
                self._stream.write(data)

    def __del__(self):
        self.close()



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


# logger using PyPubSub to improve flexibility

import inspect
import sys
from types import FrameType, TracebackType

from .pub_sub import send_message
from .topic import construct_topic, extract_topic
from .severity import TRACE, DEBUG, VERBOSE, INFO, SUCCESS, WARNING, ERROR, CRITICAL
from .subject import Subject

__all__ = ['Logger']


class Logger:
    """Main Logger class. Can be inherited from, but should be sufficient for most situations.

    :param name: Logger name. When this is an empty string (the default), it signifies this as
        the ROOT logger. Providing a ``name`` will use it as (sub-)topic, except when the ``name``
        starts with a dot (```.``), then the ``name`` (excluding the dot) will be considered a separate root.
    :type name: str, optional
    """

    __excluded_source_files = set()

    def __init__(self, name: str = "", **kwargs) -> None:
        self.name, self.topic = construct_topic(name, self.__class__.__name__)
        self.severity = 0

    def set_severity(self, severity: int) -> None:
        self.severity = severity

    @classmethod
    def register_excluded_source_file(cls, file_name: str) -> None:
        """This must be called in the source file of any descendent to exclude the respective
        entries in the call-stack.

        Recommendation: Call on source file level (module)

        .. code:: python

           Logger.register_excluded_source_file(inspect.currentframe().f_code.co_filename)

        :param file_name: Source file name
        :type file_name: str
        """
        cls.__excluded_source_files.add(file_name)

    @classmethod
    def get_excluded_source_files(cls) -> set[str]:
        """Get a copy of excluded_source_files when identifying 'real' caller in call-stack

        :rtype: set[str]
        """
        return cls.__excluded_source_files.copy()

    def critical(self, *args: str, message_id: str = None, **kwargs) -> None:
        """Log a **CRITICAL** entry.

        ``CRITICAL`` entries will include the respective source section and local variables.

        :param args: The Message(s). Every message will create a separate entry.
        :type args: str
        :param message_id: Message ID.
        :type message_id: str
        """
        self.log(CRITICAL, message_id=message_id, *args, **kwargs)

    def error(self, *args: str, message_id: str = None, **kwargs) -> None:
        """Log an **ERROR** entry.

        :param args: The Message(s). Every message will create a separate entry.
        :type args: str
        :param message_id: Message ID.
        :type message_id: str
        """
        self.log(ERROR, message_id=message_id, *args, **kwargs)

    def warning(self, *args: str, message_id: str = None, **kwargs) -> None:
        """Log a **WARNING** entry.

        :param args: The Message(s). Every message will create a separate entry.
        :type args: str
        :param message_id: Message ID.
        :type message_id: str
        """
        self.log(WARNING, message_id=message_id, *args, **kwargs)

    def success(self, *args: str, message_id: str = None, **kwargs) -> None:
        """Log a **SUCCESS** entry.

        This has been added to support responses from SAP.

        :param args: The Message(s). Every message will create a separate entry.
        :type args: str
        :param message_id: Message ID.
        :type message_id: str
        """
        self.log(SUCCESS, message_id=message_id, *args, **kwargs)

    def info(self, *args: str, message_id: str = None, **kwargs) -> None:
        """Log an **INFO** entry.

        :param args: The Message(s). Every message will create a separate entry.
        :type args: str
        :param message_id: Message ID.
        :type message_id: str
        """
        self.log(INFO, message_id=message_id, *args, **kwargs)

    def verbose(self, *args: str, message_id: str = None, **kwargs) -> None:
        """Log a **VERBOSE** entry.

        :param args: The Message(s). Every message will create a separate entry.
        :type args: str
        :param message_id: Message ID.
        :type message_id: str
        """
        self.log(VERBOSE, message_id=message_id, *args, **kwargs)

    def debug(self, *args: str, message_id: str = None, **kwargs) -> None:
        """Log a **DEBUG** entry.

        :param args: The Message(s). Every message will create a separate entry.
        :type args: str
        :param message_id: Message ID.
        :type message_id: str
        """
        self.log(DEBUG, message_id=message_id, *args, **kwargs)

    def trace(self, *args: str, message_id: str = None, **kwargs) -> None:
        """Log a **TRACE** entry.

        :param args: The Message(s). Every message will create a separate entry.
        :type args: str
        :param message_id: Message ID.
        :type message_id: str
        """
        self.log(TRACE, message_id=message_id, *args, **kwargs)

    def log(self, severity, *args: str, message_id: str = None, **kwargs) -> None:
        """Log an entry.

        While this method has been made public, be careful when using integers directly as that
        can cause issues when the interface changes.

        :param severity: The severity level of the log entry.
        :type severity: int
        :param args: The Message(s). Every message will create a separate entry.
        :type args: str
        :param message_id: Message ID.
        :type message_id: str
        """
        if severity < self.severity:
            return
        frame = self._get_caller()
        for message in args:
            self._send_log(
                Subject(
                    topic=extract_topic(self.topic),
                    severity=severity,
                    frame=frame,
                    message=str(message),
                    message_id=message_id,
                    **kwargs,
                )
            )

    def set_handler_severity(self, severity: int) -> None:
        """Set the minimum severity for all Handlers listening to the logger's topic.

        :param severity: The Message(s). Every message will create a separate entry.
        :type severity: int
        """
        # noinspection PyTypeChecker
        self._send_log(
            Subject(
                topic=extract_topic(self.topic),
                severity=severity,
                frame=None,
                message=None,
                message_id=None,
                set_severity=severity,
            )
        )

    def _get_caller(self) -> FrameType:
        """Return 'real' caller.

        If this method is overridden, make sure that the right frames are excluded.

        :meta public:
        """
        frame = inspect.currentframe()
        while frame.f_code.co_filename in self.get_excluded_source_files():
            frame = frame.f_back
        return frame

    def _send_log(self, subject: Subject) -> None:
        send_message(self.topic, subject=subject)

    def __repr__(self):
        return f"Logger(name='{self.name}')"


Logger.register_excluded_source_file(inspect.currentframe().f_code.co_filename)


def log_exception(exception_class: type, exception: Exception, trace_back: TracebackType):
    """Hook to handle uncaught exceptions.

    The entry is sent to **all** Handlers.
    A :class:`imuthes.logging.Handler` **must** implement ``handle_exception`` when it should react to this.
    """
    send_message("-", subject=Subject(topic="-", severity=sys.maxsize, frame=trace_back, message=str(exception)))


sys.excepthook = log_exception



#  hakisto - logging reimagined
#
#  Copyright (C) 2024  Bernhard Radermacher
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime
import pathlib
from types import MappingProxyType

from .indented_mixin import IndentedMixin
from .handler import Handler
from .stream import Stream

KILOBYTE = 1024
MEGABYTE = KILOBYTE * 1024
GIGABYTE = MEGABYTE * 1024

SIZE_FACTORS = MappingProxyType(dict(k=KILOBYTE, m=MEGABYTE, g=GIGABYTE))

__all__ = ['FileHandler', 'rotate_file']


def rotate_file(path: pathlib.Path, generations: int = 0) -> None:
    """'Rotates' a file, i.e. creates a backup while limiting the number of backups kept.

    The backup file are named depending on the value of ``generation``.

    Given a name of ``main.log`` the following applies.

    .. list-table::
       :header-rows: 1

       * - Condition
         - Result
         - Example
       * - value > 99
         - main.<nnn>.log
         - main.001.log
       * - 10 <= value < 100
         - main.<nn>.log
         - main.01.log
       * - 1 <= value < 10
         - main.<n>.log
         - main.1.log
       * - value = 0
         - main.<log>.bak
         - main.log.bak
       * - value < 0
         - main.<date_time>.log
         - main.060102_220405.log

    where ``date_time`` is formatted as ``yymmdd_hhmmss``.

    :param path: The path of the file to be rotated
    :type path: pathlib.Path
    :param generations: The maximum number of generations to be kept

    :param path:
    :param generations:
    """
    suffix = path.suffix

    if not generations:
        if path.exists():
            path.replace(path.with_suffix(f"{suffix}.bak"))
        return

    if generations > 0:
        num_length = len(str(generations))
        for generation in reversed(range(generations)):
            temp_path = path.with_suffix(f".{generation:0{num_length}}{suffix}") if generation else path
            if temp_path.exists():
                temp_path.replace(path.with_suffix(f".{generation + 1:0{num_length}}{suffix}"))
        return

    # when generations is negative, find all backup files
    backup_files = sorted(
        path.parent.glob(f"{path.stem}.[0-9][0-9][0-9][0-9][0-9][0-9]_[0-9][0-9][0-9][0-9][0-9][0-9]{suffix}")
    )
    while len(backup_files) >= -generations:
        backup_files.pop(0).unlink()
    path.replace(path.with_suffix(f".{datetime.datetime.now(datetime.UTC).strftime('%y%m%d_%H%M%S')}{suffix}"))


class FileHandler(IndentedMixin, Handler):
    """Output to a log file.

    :param name: Handler name (used to determine topic)
    :type name: str
    :param path: Log file to use
    :type path: :class:`pathlib.Path` or str
    :param max_size: Size when file will be rotated
    :type max_size: int or str
    :param generations: Maximum Number of generations. Negative numbers use DATE_TIME instead of numbers in the file name.
    :type generations: int
    """

    def __init__(
        self,
        name: str = "",
        path: pathlib.Path | str = None,
        max_size: int | str = "2m",
        generations: int = 10,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self.stream = Stream(path)
        if isinstance(max_size, str):
            if not max_size.isdigit():
                factor = SIZE_FACTORS[max_size[-1].lower()]
                max_size = int(max_size[:-1]) * factor
            else:
                max_size = int(max_size)
        self.max_size = max_size
        self.generations = generations

    def write(self, content: str) -> None:
        """Write content to file, check if rollover is required first"""
        with self.stream:
            self.stream.write(f"{content}\n")

    def is_rollover_required(self, content) -> bool:
        try:
            return all((self.max_size, ((self.stream.path.stat().st_size + len(content)) > self.max_size)))
        except FileNotFoundError:
            return False
